/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include "stdafx.h"

#ifdef _USRDLL
#define _QUBFITS_PLUGIN_LIB_
#endif 

#include "CustomCurve.h"

// python-interpreted custom curve


static char ParseCustomEqn_py[] =
"from math import *\n"
"from string import join\n"
"import re\n"

"def MissingName(msg):\n"
" m = re.search(\"'([^']+)'\", msg)\n"
" if not m:\n"
"  raise NameError, msg\n"
" return m.group(1)\n"

"def ParseCustomEqn(eqn, xvars):\n"
" args = ['x']+xvars\n"
" valid = 0\n"
" func = lambda : 0.0\n"

" while not valid:\n"
"  try:\n"
"   func = eval( 'lambda ' + join(args, ',') + ': ' + eqn )\n"
"   fval = float( apply( func, map(lambda z:1.0, range(len(args))) ) )\n"
"   valid = 1\n"
"  except NameError, msg:\n"
"   args.append( MissingName(str(msg)) )\n"

" return func, args[len(xvars)+1:]\n";


PyObject* py_Parse = NULL;
// PyThreadState* py_Thread = NULL;

string g_initErrMsg;

int InitCustomEqn()
{
	int result = 0;

	bool ownPython = ! Py_IsInitialized();
	if ( ownPython ) {
		Py_Initialize();
		PyEval_InitThreads();
		PyRun_SimpleString("import sys");
		PyRun_SimpleString("sys.argv = ['qubfits']");
	}

	// at this point we should have GIL because either we started it
	// or the dll was loaded holding GIL.
	// store the thread state so we can restore later
	//py_Thread       = PyThreadState_Get();

	PyObject* pmMain = PyImport_AddModule("__main__");
	PyObject* globals = PyModule_GetDict(pmMain);
	PyRun_SimpleString(ParseCustomEqn_py);
	if ( PyErr_Occurred() ) {
		PyObject *err_type, *err_val, *err_trace;
		PyErr_Fetch(&err_type, &err_val, &err_trace);
		PyObject *err_str = PyObject_Str(err_val);
		g_initErrMsg = PyString_AsString(err_str);
		PyErr_Clear();
		Py_DECREF(err_str);
		result = -1;
	} else {
		py_Parse = PyRun_String("ParseCustomEqn", Py_eval_input, globals, globals);
		if ( PyErr_Occurred() ) {
			PyObject *err_type, *err_val, *err_trace;
			PyErr_Fetch(&err_type, &err_val, &err_trace);
			PyObject *err_str = PyObject_Str(err_val);
			g_initErrMsg = PyString_AsString(err_str);
			PyErr_Clear();
			Py_DECREF(err_str);
			result = -1;
		}
	}

	if ( ownPython ) {
		PyThreadState_Swap(NULL);
		PyEval_ReleaseLock();
	}
	
	return result;
}

bool ParseCustomEqn(string eqn, vector<string>& xvars, PyObject** func, vector<string>& params, string& msg)
{
	if ( ! py_Parse ) {
		msg = g_initErrMsg;
		return false;
	}

	bool success = true;

    PyGILState_STATE state = PyGILState_Ensure();

	PyObject* args = PyTuple_New(2);
	PyTuple_SetItem(args, 0, PyString_FromString(eqn.c_str()));
	PyObject* pxvars = PyList_New(xvars.size());
	PyTuple_SetItem(args, 1, pxvars);
	for (int i=0; i<xvars.size(); ++i)
		PyList_SetItem(pxvars, i, PyString_FromString(xvars[i].c_str()));
	PyObject* py_result = PyObject_Call(py_Parse, args, NULL);
	if ( ! py_result ) {
		PyObject *err_type, *err_val, *err_trace;
		PyErr_Fetch(&err_type, &err_val, &err_trace);
		PyObject *err_str = PyObject_Str(err_val);
		msg = PyString_AsString(err_str);
		PyErr_Clear();
		Py_DECREF(err_str);
		success = false;
	} else {
		*func = PySequence_GetItem(py_result, 0);
		PyObject* py_params = PySequence_GetItem(py_result, 1);
		int n = PySequence_Length(py_params);
		for (int i=0; i<n; ++i) {
			PyObject* py_param = PySequence_GetItem(py_params, i);
			params.push_back(PyString_AsString(py_param));
			Py_DECREF(py_param);
		}

		Py_DECREF(py_params);
		Py_DECREF(py_result);
	}
	Py_DECREF(args);

	  PyGILState_Release(state);

	return success;
}

QUB_Curve_Custom::QUB_Curve_Custom()
: state( QUB_Tree::Create("Curve") ),
  ncomp( 1 ),
  func( NULL )
{
	state["Name"].setData(getName());
	state["ComponentCount"].setData(QTR_TYPE_INT, ncomp);
	setEqn("m*x + b");
}
void __cdecl QUB_Curve_Custom::Free()
{
	delete this;
}
IQUB_Evts * __cdecl QUB_Curve_Custom::Events()
{
	return &events;
}
void __cdecl QUB_Curve_Custom::setExtraVars(char **names, int count)
{
	extraVars.clear();
	for (int i=0; i<count; ++i)
		extraVars.push_back( names[i] );
	for (int i=0; i<count; ++i) {
		size_t sep = extraVars[i].find(' ');
		sep = min(sep, extraVars[i].find('-'));
		if (sep != string::npos)
			extraVars[i] = extraVars[i].substr(0, sep);
	}
	setEqn(getEqn()); // re-parse eqn with different xtra vars
}
const char * __cdecl QUB_Curve_Custom::getName()
{
	return "Custom";
}
const char * __cdecl QUB_Curve_Custom::getEqn()
{
	return eqn.c_str();
}
void __cdecl QUB_Curve_Custom::setEqn(const char *x)
{
	PyObject* newFunc;
	vector<string> newParams;
	if ( ! ParseCustomEqn(x, extraVars, &newFunc, newParams, message) ) {
		events.send("Curve Msg", this);
		return;
	}

	QUB_Tree bkState = state.clone();
	bkState.removeChild(bkState["Eqn"]);
	ncomp = 0;

	func = newFunc;
	eqn = x;
	state["Eqn"].setData(eqn);
	paramNames = newParams;
	fullParamNames = newParams;
	paramVals = vector<double>(newParams.size(), 1.0);
	paramConst = vector<int>(newParams.size(), 0);

	updateState(bkState.getImpl());
	events.send("Curve Eqn", this);
}
int __cdecl QUB_Curve_Custom::canSetEqn()
{
	return 1;
}
int __cdecl QUB_Curve_Custom::getComponentCount()
{
	return ncomp;
}
void __cdecl QUB_Curve_Custom::setComponentCount(int x)
{
	QUB_Tree components = state["Components"];
	for (; ncomp>x; --ncomp)
		components.removeChild(stateComponents[ncomp-1]);
	stateComponents.resize(x);

	int npar = paramNames.size();
	paramVals.resize(npar * x);
	fullParamNames.resize(npar * x);
	paramConst.resize(npar * x);
	for ( ; ncomp<x; ++ncomp) {
		for (int p=0; p<npar; ++p) {
			paramVals[ncomp*npar+p] = 1.0;
			if ( x > 1 ) {
				ostringstream ost;
				ost << paramNames[p] << (ncomp+1);
				fullParamNames[ncomp*npar+p] = ost.str();
			} else {
				fullParamNames[ncomp*npar+p] = paramNames[p];
			}
		}
		if ( stateComponents[ncomp].isNull() ) {
			stateComponents[ncomp] = components.appendChild("Component");
			for (int p=0; p<npar; ++p) {
				QUB_Tree qpar = stateComponents[ncomp][paramNames[p]];
				qpar.setNumData(QTR_TYPE_DOUBLE, 2, 1, (void*)0);
				qpar.dataAs(0,0,(double)0) = paramVals[ncomp*npar+p];
				qpar.dataAs(1,0,(double)0) = paramConst[ncomp*npar+p];
			}
		}
	}

	state["ComponentCount"].setData(QTR_TYPE_INT, ncomp);
	events.send("Curve CompCount", this);
	events.send("Curve Params", this);
}
int __cdecl QUB_Curve_Custom::getParamCount()
{
	return paramVals.size();
}
const char * __cdecl QUB_Curve_Custom::getParamName(int ix)
{
	return fullParamNames[ix].c_str();
}
double __cdecl QUB_Curve_Custom::getParamVal(int ix)
{
	return paramVals[ix];
}
void __cdecl QUB_Curve_Custom::setParamVal(int ix, double val, int sendEvent)
{
	paramVals[ix] = val;

	int npar = paramNames.size();
	int c = ix / npar;
	int p = ix % npar;
	if ( stateComponents[c][paramNames[p]].dataCount() < 2 )
		stateComponents[c][paramNames[p]].setNumData(QTR_TYPE_DOUBLE, 2, 1, (void*)0);
	stateComponents[c][paramNames[p]].dataAs(0,0,(double)0) = val;
	if ( sendEvent )
		events.send("Curve ParamVal", this);
}
int __cdecl QUB_Curve_Custom::getParamConst(int ix)
{
	return paramConst[ix];
}
void __cdecl QUB_Curve_Custom::setParamConst(int ix, int isConst)
{
	paramConst[ix] = isConst;
	int npar = paramNames.size();
	QUB_Tree qpar = stateComponents[ix / npar][paramNames[ix % npar]];
	if ( qpar.dataCount() == 0 )
		qpar.setNumData(QTR_TYPE_DOUBLE, 2, 1, (void*)0);
	else if ( qpar.dataCount() == 1 )
		qpar.resizeData(2);
	qpar.dataAs(1,0,(double)0) = (double) isConst;
	events.send("Curve ParamConst", this);
}
void __cdecl QUB_Curve_Custom::initParams(double *xx, double *yy, int n)
{
	//
}
void __cdecl QUB_Curve_Custom::evalArrayAll(QF_Data *data, double *yy)
{
	int n = data->n;
	double *xx = data->xx;
	for (int i=0; i<n; ++i)
		yy[i] = 0.0;

	if ( ! func )
		return;

	vector<double*> vvv_;
	for (int i=0; i<extraVars.size(); ++i) {
		// find vi | 0<=vi<data->nv, data->vname[vi] == extraVars[i]
		string& vname = extraVars[i];
		int vi;
		for (vi=0; vi<data->nv; ++vi)
			if ( vname == data->vname[vi] )
				break;
		vvv_.push_back( (vi<data->nv) ? data->vvv[vi] : NULL );
	}
	double **vvv = vvv_.size() ? &(vvv_[0]) : NULL;

	  //BeginPython();
        PyGILState_STATE state = PyGILState_Ensure();

	int npar = paramNames.size();
	int nextra = extraVars.size();
	PyObject* args = PyTuple_New(1+nextra+npar);
	
	for (int c=0; c<ncomp; ++c) {
		for (int p=0; p<npar; ++p)
			PyTuple_SetItem(args, nextra+p+1, PyFloat_FromDouble(paramVals[c*npar+p]));
		for (int i=0; i<n; ++i) {
			PyTuple_SetItem(args, 0, PyFloat_FromDouble(xx[i]));
			for (int j=0; j<nextra; ++j)
				PyTuple_SetItem(args, j+1, PyFloat_FromDouble(vvv[j] ? vvv[j][i] : 0.0));
			PyObject *result = PyObject_Call(func, args, NULL);
			if ( PyErr_Occurred() )
				PyErr_Clear(); // should report?!
			else {
				yy[i] += PyFloat_AsDouble(result);
				Py_DECREF(result);
			}
		}
	}

	Py_DECREF(args);

	  //EndPython();
	  PyGILState_Release(state);
}
void __cdecl QUB_Curve_Custom::evalArrayOne(QF_Data *data, double *yy, int c)
{
	int n = data->n;
	double *xx = data->xx;
	for (int i=0; i<n; ++i)
		yy[i] = 0.0;

	if ( ! func )
		return;

	vector<double*> vvv_;
	for (int i=0; i<extraVars.size(); ++i) {
		// find vi | 0<=vi<data->nv, data->vname[vi] == extraVars[i]
		string& vname = extraVars[i];
		int vi;
		for (vi=0; vi<data->nv; ++vi)
			if ( vname == data->vname[vi] )
				break;
		vvv_.push_back( (vi<data->nv) ? data->vvv[vi] : NULL );
	}
	double **vvv = &(vvv_[0]);

        PyGILState_STATE state = PyGILState_Ensure();
	//BeginPython();

	int npar = paramNames.size();
	int nextra = extraVars.size();
	PyObject* args = PyTuple_New(1+nextra+npar);
	
	//for (int c=0; c<ncomp; ++c) {
		for (int p=0; p<npar; ++p)
			PyTuple_SetItem(args, nextra+p+1, PyFloat_FromDouble(paramVals[c*npar+p]));
		for (int i=0; i<n; ++i) {
			PyTuple_SetItem(args, 0, PyFloat_FromDouble(xx[i]));
			for (int j=0; j<nextra; ++j)
				PyTuple_SetItem(args, j+1, PyFloat_FromDouble(vvv[j] ? vvv[j][i] : 0.0));
			PyObject *result = PyObject_Call(func, args, NULL);
			if ( PyErr_Occurred() )
				PyErr_Clear(); // should report?!
			else {
				yy[i] += PyFloat_AsDouble(result);
				Py_DECREF(result);
			}
		}
	//}

	Py_DECREF(args);

	  //EndPython();
	  PyGILState_Release(state);
}
int __cdecl QUB_Curve_Custom::hasDeriv()
{
	return 0;
}
void __cdecl QUB_Curve_Custom::evalDerivArray(QF_Data *data, double *dpar)
{}
const char * __cdecl QUB_Curve_Custom::getMessage()
{
	return message.c_str();
}
QTR_Impl* __cdecl QUB_Curve_Custom::getState()
{
	QTR_Impl* pstate = state.getImpl();
	QTR_INCREF(pstate);
	return pstate;
}
void __cdecl QUB_Curve_Custom::updateState(QTR_Impl* pustate)
{
	QUB_Tree ustate(pustate);
	QUB_Tree sub;
	if ( ustate["Name"].dataAsString() == this->getName() ) {
		sub = * ustate.find("Eqn");
		if ( sub.dataCount() > 0 )
			this->setEqn(sub.dataAsString().c_str());
		sub = * ustate.find("ComponentCount");
		if ( sub.dataCount() > 0 )
			this->setComponentCount( sub.dataAsInt(1) );
		int ipar = 0;
		for ( QUB_TreeMonoIter ci = ustate["Components"].find("Component");
			! ci->isNull(); ci.nextSameName() ) {
		  for (vector<string>::iterator pni = paramNames.begin(); pni != paramNames.end(); ++pni) {
			sub = * ci->find(*pni);
			if ( sub.dataCount() > 0 )
				this->setParamVal(ipar++, sub.dataAs(0,0,(double)0), 0);
			if ( sub.dataCount() > 1 )
				this->setParamConst(ipar-1, (int)sub.dataAs(1,0,(double)0));
		  }
		}
		if ( this->getParamCount() > 0 )
			this->setParamVal(0, this->getParamVal(0), 1); // trigger event
	}
}


IQUB_Curve* __stdcall CreateCustomCurve()
{
	return new QUB_Curve_Custom;
}

